from jutil.database import mysql_database_singleton, database_controller

from ScopusWp.extend import Config

from ScopusWp.scopus.config import DB_NAME

from sqlalchemy import Column, Integer, DATETIME, String

import datetime


class ScopusMySQLDatabase(mysql_database_singleton(
    Config.get_instance()['MYSQL']['username'],
    Config.get_instance()['MYSQL']['password'],
    Config.get_instance()['MYSQL']['host'],
    DB_NAME
)):
    pass


BASE = ScopusMySQLDatabase.get_base()


class BlackListing(BASE):

    __tablename__ = 'blacklist'

    scopusID = Column(Integer, primary_key=True)
    description = Column(String(200))
    created = Column(DATETIME)


class DatabaseController(database_controller(ScopusMySQLDatabase)):

    def __init__(self):
        super().__init__()

    def get_blacklist(self):
        return list(map(
            lambda black_listing: int(black_listing.scopusID),
            self._select_all(BlackListing)
        ))

    def add_blacklist(self, scopus_id, description=''):
        black_listing = BlackListing(
            scopusID=scopus_id,
            description=description,
            created=datetime.datetime.now()
        )
        self._add_blacklisting(black_listing)

    def _add_blacklisting(self, black_listing):
        self._add(
            BlackListing,
            black_listing,
            lambda x: BlackListing.scopusID == x.scopusID
        )



from sqlalchemy import Column, ForeignKey, Integer, String, Text, DATETIME, BigInteger, Table, and_
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, Session

from ScopusWp.database import MySQLDatabase, get_or_create
from ScopusWp.util import IDManager

from jutil.database import sqlalchemy_many_list_tuple, sqlalchemy_add_children, session_refresh
from jutil.processing import DictQuery

from ScopusWp.dat import GeneralPublication, PublicationAuthor, PublicationJournal, PublicationLink, PublicationOrigin
import datetime

from ScopusWp.database import BASE

from pprint import pprint


publication_author_junction = Table(
    'publication_author_junction',
    BASE.metadata,
    Column('authorID', BigInteger, ForeignKey('author.authorID'), primary_key=True),
    Column('publicationID', BigInteger, ForeignKey('publication.publicationID'), primary_key=True)
)

publication_tag_junction = Table(
    'publication_tag_junction',
    BASE.metadata,
    Column('publicationID', BigInteger, ForeignKey('publication.publicationID'), primary_key=True),
    Column('tagID', BigInteger, ForeignKey('tag.tagID'), primary_key=True)
)

publication_category_junction = Table(
    'publication_category_junction',
    BASE.metadata,
    Column('publicationID', BigInteger, ForeignKey('publication.publicationID'), primary_key=True),
    Column('CategoryID', BigInteger, ForeignKey('category.categoryID'), primary_key=True)
)

publication_citation_junction = Table(
    'publication_citation_junction',
    BASE.metadata,
    Column('publicationID', BigInteger, ForeignKey('publication.publicationID'), primary_key=True),
    Column('citationID', BigInteger, ForeignKey('publication.publicationID'), primary_key=True)
)


class Publication(BASE):

    __tablename__ = 'publication'

    publicationID = Column(BigInteger, primary_key=True)
    doi = Column(String(200))
    title = Column(String(500))
    description = Column(Text)
    published = Column(DATETIME)
    originID = Column(BigInteger, ForeignKey('origin.originID'))
    journalID = Column(BigInteger, ForeignKey('journal.journalID'))

    origin = relationship(
        'Origin',
        uselist=False,
        back_populates='publication'
    )
    journal = relationship(
        'Journal',
        uselist=False,
        back_populates='publications'
    )
    links = relationship(
        'Link',
        back_populates='publication'
    )
    authors = relationship(
        "Author",
        secondary=publication_author_junction,
        back_populates='publications'
    )
    categories = relationship(
        'Category',
        secondary=publication_category_junction,
        back_populates='publications'
    )
    tags = relationship(
        'Tag',
        secondary=publication_tag_junction,
        back_populates='publications'
    )
    citedby = relationship(
        'Publication',
        secondary=publication_citation_junction,
        primaryjoin='Publication.publicationID == publication_citation_junction.c.publicationID',
        secondaryjoin='Publication.publicationID == publication_citation_junction.c.citationID'
    )

    def add_citation(self, publication):
        if publication not in self.citedby:
            self.citedby.append(publication)


class Author(BASE):

    __tablename__ = 'author'

    authorID = Column(BigInteger, index=True, primary_key=True)
    first_name = Column(String(500))
    last_name = Column(String(500))

    publications = relationship(
        'Publication',
        secondary=publication_author_junction,
        back_populates='authors'
    )


class Tag(BASE):

    __tablename__ = 'tag'

    tagID = Column(BigInteger, primary_key=True)
    content = Column(String(500))

    publications = relationship(
        'Publication',
        secondary=publication_tag_junction,
        back_populates='tags'
    )


class Journal(BASE):

    __tablename__ = 'journal'

    journalID = Column(BigInteger, primary_key=True)
    name = Column(String(200))
    volume = Column(String(200))

    publications = relationship(
        'Publication',
        back_populates='journal'
    )


class Category(BASE):

    __tablename__ = 'category'

    categoryID = Column(BigInteger, primary_key=True)
    content = Column(String(500))

    publications = relationship(
        'Publication',
        secondary=publication_category_junction,
        back_populates='categories'
    )


class Origin(BASE):

    __tablename__ = 'origin'

    originID = Column(BigInteger, primary_key=True)
    id = Column(String(100))
    typeID = Column(Integer, ForeignKey('type.typeID'))
    text = Column(Text)

    publication = relationship(
        'Publication',
        back_populates='origin'
    )
    type = relationship(
        'OriginType',
        back_populates='origins'
    )


class Link(BASE):

    __tablename__ = 'link'

    linkID = Column(BigInteger, primary_key=True)
    name = Column(String(200))
    href = Column(String(200))
    publicationID = Column(BigInteger, ForeignKey('publication.publicationID'))

    publication = relationship(
        'Publication',
        back_populates='links'
    )


class OriginType(BASE):

    __tablename__ = 'type'

    typeID = Column(Integer, primary_key=True)
    name = Column(String(100))

    origins = relationship(
        'Origin',
        back_populates='type'
    )


class PublicationRegister:

    def __init__(self):

        """
        self.id_managers = {
            'origin': IDManager('origin'),
            'publication': IDManager('publication'),
            'author': IDManager('author'),
            'link': IDManager('link'),
            'tag': IDManager('tag'),
            'category': IDManager('category'),
        }
        """

        self.session = MySQLDatabase.get_session()

        self.publication_id_manager = IDManager('general_publication')
        self.remaining_citations = {}

        self.publication_processor = PublicationInserter(self)

    @session_refresh(MySQLDatabase)
    def insert(self, publication_dict):

        self.publication_processor.set(publication_dict)
        self.publication_processor.set_session(self.session)
        publication, remaining_citations = self.publication_processor.process()
        self.remaining_citations.update(remaining_citations)

        self.session.add(publication)
        self.session.commit()

    @session_refresh(MySQLDatabase)
    def contains_publication(self, publication_dict):
        query = self.session.query(Publication)\
                    .join(Publication.origin, aliased=True).filter_by(id=publication_dict['origin']['id'])\
                    .join(Origin.type).filter_by(name=publication_dict['origin']['name'])
        return len(list(query)) != 0

    @session_refresh(MySQLDatabase)
    def select(self, publication_id):
        publication = self.session.query(Publication).get(publication_id)
        return GeneralPublication.from_model(publication)

    @session_refresh(MySQLDatabase)
    def select_origin(self, name, text=None):
        if text is None:
            query = self.session.query(Publication)\
                        .join(Publication.origin)\
                        .join(OriginType).filter_by(name=name)
        else:
            query = self.session.query(Publication) \
                .join(Publication.origin).filter_by(text=text) \
                .join(OriginType).filter_by(name=name)
        publications = list(query)
        return list(map(lambda x: GeneralPublication.from_model(x), publications))

    @session_refresh(MySQLDatabase)
    def select_origin_id(self, name, text=None):
        if text is None:
            query = self.session.query(Publication)\
                        .join(Publication.origin)\
                        .join(OriginType).filter_by(name=name)
        else:
            query = self.session.query(Publication) \
                .join(Publication.origin).filter_by(text=text) \
                .join(OriginType).filter_by(name=name)
        return list(map(lambda x: int(x.origin.id), query))

    def query(self, model_class, filter_statement):
        return list(self.session.query(model_class).filter(filter_statement))

    def wipe(self):
        self.session.query(Link).delete()
        self.session.execute('TRUNCATE TABLE publication_author_junction')
        self.session.execute('TRUNCATE TABLE publication_category_junction')
        self.session.execute('TRUNCATE TABLE publication_tag_junction')
        self.session.execute('TRUNCATE TABLE publication_citation_junction')
        self.session.query(Publication).delete()
        self.session.query(Tag).delete()
        self.session.query(Category).delete()
        self.session.query(Author).delete()
        self.session.query(Origin).delete()
        self.session.query(Journal).delete()
        self.session.commit()


class PublicationInserter(DictQuery):

    def __init__(self, register):
        DictQuery.__init__(self)
        self.session = register.session
        self.id_manager = register.publication_id_manager
        self.publication = None

        self.remaining_citations = {}

    def set_session(self, session):
        self.session = session

    def query_exception(self, query, query_dict, default):
        return default

    def process(self):
        self.publication = Publication(
            publicationID=self.id_manager.new(),
            doi=self.query_dict('doi', ''),
            title=self.query_dict('title', ''),
            description=self.query_dict('description', ''),
            published=self.query_dict('published', datetime.datetime.now())
        )

        self.publication.origin = self.origin()
        self.publication.journal = self.journal()
        self.publication.links = self.links()
        sqlalchemy_add_children(self.publication.tags, self.tags())
        sqlalchemy_add_children(self.publication.authors, self.authors())
        sqlalchemy_add_children(self.publication.categories, self.categories())
        for publication in self.citations():
            self.publication.add_citation(publication)

        self.id_manager.save()

        return self.publication, self.remaining_citations

    def origin(self):
        # Getting the origin type
        origin_type = self.session.query(OriginType).filter(
            OriginType.name == self.query_dict('origin/name', '')
        ).first()
        return Origin(
            id=self.query_dict('origin/id', 0),
            text=self.query_dict('origin/text', ''),
            type=origin_type
        )

    def journal(self):

        journal_tuple = (
            self.query_dict('journal', ''),
            self.query_dict('volume', '')
        )

        # The many list tuple function returns a tuple, of which the first item is a list with all the model objects
        # corresponding to the parameter list given (in this cae only the one constructed tuple) either by getting
        # them from the db or constructing a new Model object
        return sqlalchemy_many_list_tuple(
            self.session,
            Journal,
            [journal_tuple],
            lambda x: and_(Journal.name == x[0], Journal.volume == x[1]),
            lambda x: Journal(name=x[0], volume=x[1])
        )[0][0]  # The corresponding model object to the one constructed above must be the first item of this first list

    def links(self):
        # The links are unique to each publication, which means they all have to be constructed from scratch.
        # Each item of the links list is a dict again with the info about the link name and the actual url
        return list(map(
            lambda link: Link(
                name=link['name'],
                href=link['href']
            ),
            self.query_dict('links', [])
        ))

    def authors(self):
        authors, authors_remaining = sqlalchemy_many_list_tuple(
            self.session,
            Author,
            self.query_dict('authors', []),
            lambda author: and_(
                Author.first_name == author['first_name'],
                Author.last_name == author['last_name'],
                Author.authorID == author['id']
            ),
            lambda author: Author(
                first_name=author['first_name'],
                last_name=author['last_name']
            )
        )
        return authors

    def tags(self):
        tags, tags_remaining = sqlalchemy_many_list_tuple(
            self.session,
            Tag,
            self.query_dict('tags', []),
            lambda tag: Tag.content == tag,
            lambda tag: Tag(content=tag)
        )
        return tags

    def categories(self):
        categories, categories_remaining = sqlalchemy_many_list_tuple(
            self.session,
            Category,
            self.query_dict('categories', []),
            lambda category: Category.content == category,
            lambda category: Category(content=category)
        )
        return categories

    def citations(self):
        citations, remaining_publications = sqlalchemy_many_list_tuple(
            self.session,
            Publication,
            self.query_dict('citations', []),
            lambda origin_dict: self.citations_filter_function(origin_dict),
            lambda origin_dict: {(origin_dict['id'], origin_dict['name']): self.publication.publicationID}
        )
        for item in remaining_publications:
            self.remaining_citations.update(item)
        return list(filter(lambda x: not isinstance(x, dict), citations))

    @staticmethod
    def citations_filter_function(origin_dict):
        return and_(
            Publication.origin.has(id=origin_dict['id']),
            Origin.type.has(name=origin_dict['name'])
        )


if __name__ == '__main__':
    r = PublicationRegister()
    BASE.metadata.drop_all(MySQLDatabase._engine)
    MySQLDatabase.create_database(BASE)




from setuptools import setup


setup(
    name='scopus.wp',
    version='0.3.0.70',
    description='A tool for a wordpress server which will automatically post science publications from scopus database',
    url='https://github.com/the16thpythonist/ScopusWp',
    author='Jonas Teufel',
    author_email='jonseb1998@gmail.com',
    license='MIT',
    packages=['ScopusWp', 'ScopusWp/scopus', 'ScopusWp/wordpress'],
    include_package_data=True,
    install_requires=[
        'requests>=2.0',
        'mysqlclient>=1.2',
        'unidecode>=0.4',
        'tabulate>=0.8',
        'python-wordpress-xmlrpc>=2.3',
        'jinja2>=2.10',
        'sqlalchemy>=1.2',
        'memory_profiler',
        'jutily'
    ],
    python_requires='>=3, <4',
    package_data={
        '': []
    },
)



from ScopusWp.dat import GeneralPublication

from ScopusWp.wordpress.config import PATH

from jinja2 import Template

import pathlib


class PublicationCommentView:

    def __init__(self, general_publication):
        self.publication = general_publication  # type: GeneralPublication
        self.templates_path = pathlib.Path(PATH) / 'templates'

    def date(self):
        return self.publication.published

    def content(self):
        with (self.templates_path / 'comment.html').open(mode='r') as file:
            template_string = file.read()
            template = Template(template_string)
            a = template.render(
                authors=self.publication.authors,
                title=self.publication.title,
                journal=self.publication.journal,
                volume=self.publication.volume,
                year=self.publication.published.year,
            )
            return a


class PublicationPostView:

    def __init__(self, general_publication):
        self.publication = general_publication  # type: GeneralPublication
        self.templates_path = pathlib.Path(PATH) / 'templates'
        self.author_limit = 20

    def title(self):
        return self.publication.title.encode('utf-8')

    def date(self):
        return self.publication.published

    def slug(self):
        return str(self.publication.id)

    def content(self):
        with (self.templates_path / 'post.html').open(mode='r') as file:
            template_string = file.read()
            template = Template(template_string)
            return template.render(
                authors=self._authors(),
                journal=self.publication.journal,
                volume=self.publication.volume,
                year=self.publication.published.year,
                doi=self.publication.doi,
                description=self.publication.description,
                links=self._links()
            )

    def excerpt(self):
        return u''

    def tags(self):
        return self.publication.tags

    def categories(self):
        return self.publication.categories

    def _authors(self):
        author_list = []
        for author in self.publication.authors:
            try:
                author_list.append(author.index_name)
            # In case there is a index error, that means the first name of the author does not even contain a single
            # character in which case it is useless to even add this string to the list of authors
            except IndexError:
                continue
        if len(author_list) >= self.author_limit:
            author_list = author_list[:self.author_limit]
        return ', '.join(author_list)

    def _links(self):
        with (self.templates_path / 'button.html').open(mode='r') as file:
            template = Template(file.read())
            return template.render(
                doi=self.publication.doi
            )



from ScopusWp.config import Config
from ScopusWp.config import SQL_LOGGING_EXTENSION

import logging

import MySQLdb

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Session

from sqlalchemy.ext.declarative import declarative_base

from ScopusWp.config import Config

from jutil.database import mysql_database_singleton

BASE = declarative_base()


def get_or_create(session, model, **kwargs):
    instance = session.query(model).filter_by(**kwargs).first()
    if instance:
        return instance
    else:
        instance = model(**kwargs)
        session.add(instance)
        session.commit()
        return instance


class MySQLDatabase(mysql_database_singleton(
    Config.get_instance()['MYSQL']['username'],
    Config.get_instance()['MYSQL']['password'],
    Config.get_instance()['MYSQL']['host'],
    '__pubcontrol'
)):
    pass


######################
# DANGER LEGACY CODE #
######################


class Db:
    """
    This is the singleton class which manages the database connection object.
    On the first call to the get instance method a new "MySQLdb" connection object is being created using the database
    name, username and password given in the config file of the project.
    The connection object is stored in a class variable and is returned each time the get instance method is called
    again.
    """
    _instance = None

    def __init__(self):
        pass

    @staticmethod
    def get_instance():
        """
        This method returns the connection object for the database that is stored inside this singleton.

        :return: MySQLdb.connect() object
        """
        # In case this is the first call to the method a new connection object will be created and saved in the static
        # class variable.
        if Db._instance is None:
            Db.new_instance()

        # Otherwise the saved object will be returned
        return Db._instance

    @staticmethod
    def new_instance():
        """
        This method creates the new database connection object from the database name, the username and password in
        the config file of the project and then saves the object in the class variable
        :return:
        """
        config = Config.get_instance()
        database = config['MYSQL']['database']
        username = config['MYSQL']['username']
        password = config['MYSQL']['password']
        connector = MySQLdb.connect(host='localhost', user=username, passwd=password, db=database)
        Db._instance = connector


class SQLDatabaseAccessInterface:
    """
    Simple interface for a sql database access wrapper.
    The database wrapper, however it is implemented should implement two methods of interacting with a database:

    EXECUTE
    The execute method takes some sql code snippet and simply executes it onto the database, without any return
    whatsoever, this is designed to be used for insert or update statements

    SELECT
    The select method is supposed to also execute the sql code on the database, but return a list of tuples. This
    is manly supposed to be used with the sql select statement and the returned list to be the list of rows as the
    response to such a select statement
    """
    def save(self):
        """
        Supposed to save the changes made to the database

        :return: void
        """
        raise NotImplementedError()

    def execute(self, sql):
        """
        Supposed to execute a sql command without return

        :param sql: The string sql code to return
        :return: void
        """
        raise NotImplementedError()

    def select(self, sql):
        """
        Supposed to execute a sql command and then return a list of lists, where each sub list represents one row
        in the database.

        :param sql: The string sql command to return
        :return: A list of lists
        """
        raise NotImplementedError()


class MySQLDatabaseAccess(SQLDatabaseAccessInterface):
    """
    MySQL database access wrapper.
    THis class implements the SQL database access interface.
    it is used to execute Insert, update statements onto the database or fetch rows of data with the select statement.
    This wrapper is to be used everywhere, where a interface to the mysql database is needed.
    """
    def __init__(self):
        # Getting the database access from the singleton and creating a new cursor to interact with the database
        self.db = Db.get_instance()
        self.cursor = self.db.cursor()

        # Getting the according logger
        self.logger = logging.getLogger('MYSQLDatabaseAccess')

    def save(self):
        #self.db.commit()
        #self.cursor.close()
        #self.cursor = self.db.cursor()
        a = 1

    def execute(self, sql):
        """
        Executes the given string of sql statements on the mysql database of this project, which is specified in the
        config file.

        :param sql: The string of sql code.
        :return: void
        """
        try:
            self.cursor.execute(sql)
        except Exception as exception:
            # Logging the error
            error_string = (
                'During the handling of the SQL expression "{}" occurred the following exception "{}"'
            ).format(
                str(sql).replace('\n', ' '),
                str(exception).replace('\n', ' ')
            )
            self.logger.error(error_string)
            # Actually raising an exception
            raise exception

    def select(self, sql):
        """
        Executes the sql code string on the database and returns the list of rows selected.
        This method is supposed to be used with the select statement.

        :param sql: The sql code string.
        :return: [data tuples, each one for one row, that matches the select statement]
        """
        # First: Actually executing the sql command
        self.execute(sql)

        # Second: iterating through the contents of the cursor
        row_list = []
        for row in self.cursor:
            if isinstance(row, (list, tuple)):
                row_list.append(row)
            else:
                # If the contents of the cursor are no rows, than either there was an error, or the method was misused
                error_message = (
                    'There seems to be an error with the SQL statement {}, '
                    'the returned data is not in the row format: {}'
                ).format(
                    str(sql).replace('\n', ' '),
                    str(row).replace('\n', ' ')
                )
                self.logger.error(error_message)
        return row_list
